#include "pcie40_reg.h"
#include "pcie40_ecs.h"
#include <unistd.h>
#include <stdio.h>
#include <vector>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iostream>
// ===== 
// CALL C functions
// =====

#ifndef D
#define D(a,b,c) (((a)>>(c))&((1<<((b)+1-(c)))-1))
#define B(a,b)   D(a,b,b)
#define Bs(a,b,s)   (B(a,b)?(s):"")
#define Ds(a,b,c,s)   (D(a,b,c)?(s):"")
#endif

#define SIZE_MEMO_FPGA 1024 // maximum size of memory in FPGA

int pcie40_status(int dev){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_PCIE40_STATUS);
  return ret;
}


int pcie40_run_number(int dev){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_RUM_NUMBER);
  return (ret & 0xFF00) >> 8 | (ret & 0x00FF) << 8;
}

int pcie40_trgtag_number(int dev){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_TRGTAG_NUMBER);
  return ret;
}

int pcie40_memo_status(int dev){
  return B(pcie40_status(dev), 8);
}

int pcie40_ttd_status(int dev){
  return B(pcie40_status(dev), 12);
}

int pcie40_ttdclk_status(int dev){
  return B(pcie40_status(dev), 11);
}

int pcie40_trg_number(int dev){
  return D(pcie40_status(dev), 28, 17);
}

int pcie40_trg_type(int dev){
  return D(pcie40_status(dev), 16, 13);
}

int pcie40_rxlink_status(int dev){
  return B(pcie40_status(dev), 0);
}

int pcie40_txlink_status(int dev){
  return B(pcie40_status(dev), 1);
}

int pcie40_fee_synced(int dev){
  return B(pcie40_status(dev), 2);
}

int pcie40_disp_status(int dev){
  return B(pcie40_status(dev), 3);
}

int pcie40_rxdata_error(int dev){
  return B(pcie40_status(dev), 4);
}

//----------------------------------------------
// Belle2link dedicated
//----------------------------------------------

// Status of gbt specified by lane_up
int pcie40_gbt_status(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_LANEUP1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_LANEUP2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch - 24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch - 24 ));
  }
  return -1;
}
// Status of b2l specified by link_ok
int pcie40_b2l_status(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_STATUS1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_STATUS2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch - 24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch - 24 ));
  }
  return -1;
}
// Status of b2l include link_ok and state machine are ready
int pcie40_b2l_tiestatus(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_TIE_STATUS1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_TIE_STATUS2);  
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch -24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch - 24 ));
  }
  return -1;
}

int pcie40_b2l_mask(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MASK1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MASK2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch -24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch - 24 ));
  }
  return -1;
}
// set mask for corresponding b2l: 0->mask, 1->unmask
int pcie40_b2l_set_mask(int dev, int ch, bool mask){
  int ret = -1;
  int cur_stat = -1;
  int set_stat = -1;
  if(ch >= 0 && ch <=23){
    cur_stat = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MASK1);
    if(mask)
      set_stat = (~(1<<ch)&cur_stat) & cur_stat;
    else
      set_stat = ((1<<ch)|cur_stat);
    ret = ecs_write(dev, SLC_BAR,  SLC_B2LINK_MASK1, set_stat);
    if(ret != 0) return -1;
  }else if (ch >= 24 && ch <= 47 ){
    cur_stat = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MASK2);
    if(mask)
      set_stat = (~( 1<<(ch-24) )&cur_stat) & cur_stat;
    else
      set_stat = ( (1<<(ch-24) )|cur_stat);
    ret = ecs_write(dev, SLC_BAR,  SLC_B2LINK_MASK2, set_stat);
    if(ret != 0) return -1;
  }
  return ret;  
}

// set mask for corresponding b2l: 0->mask, 1->unmask
int pcie40_b2l_set_bankmask(int dev, int bank, bool mask){
  int ret = -1;
  int cur_stat = -1;
  int set_stat = -1;
  int ch = -1;
  if(bank >= 0 && bank <=3){
    cur_stat = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MASK1);
    for(int ich=0; ich<6; ich++){
      ch = bank*6 + ich;
      if(mask)
	set_stat = (~(1<<ch)&cur_stat) & cur_stat;
      else
	set_stat = ((1<<ch)|cur_stat);
      cur_stat = set_stat;
    }
    ret = ecs_write(dev, SLC_BAR,  SLC_B2LINK_MASK1, set_stat);
    if(ret != 0) return -1;
  }else if (bank >= 4 && bank <= 7 ){
    cur_stat = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MASK2);
    for(int ich=0; ich<6; ich++){
      ch = bank*6 + ich;
      if(mask)
	set_stat = (~( 1<<(ch-24) )&cur_stat) & cur_stat;
      else
	set_stat = ( (1<<(ch-24) )|cur_stat);
      cur_stat = set_stat;
    }
    ret = ecs_write(dev, SLC_BAR,  SLC_B2LINK_MASK2, set_stat);
    if(ret != 0) return -1;
  }
  return ret;  
}

int pcie40_b2l_txstatus(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_TX_STATUS1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_TX_STATUS2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch - 24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch - 24 ));
  }
  return -1;
}

int pcie40_b2l_rxstatus(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_RX_STATUS1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_RX_STATUS2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch - 24 ) ) ) 
           >> ( SLC_BASE_BIT + ch - 24));
  }
  return -1;
}

int pcie40_b2l_txready(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_TX_READY1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_TX_READY2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch - 24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch -24 ));
  }
  return -1;
}

int pcie40_b2l_rxready(int dev, int ch){
  int ret = -1;
  if(ch >= 0 && ch <=23){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_RX_READY1);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch ) ) ) 
	     >> ( SLC_BASE_BIT + ch ));
  }else if (ch >= 24 && ch <= 47 ){
    ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_RX_READY2);
    return ( ( ret & ( 1 << ( SLC_BASE_BIT + ch - 24 ) ) ) 
	     >> ( SLC_BASE_BIT + ch -24 ));
  }
  return -1;
}

int pcie40_b2l_cnt_vetoidle(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_VETOIDLE + ch*0x20);
  
  return ( ret & 0xFFFF);
}

int pcie40_b2l_cnt_vetodata(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_VETODATA + ch*0x20);
  
  return ( ret & 0xFFFF);
}

int pcie40_b2l_3rdrxdata(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_BUFDOWN0 + ch*0x20);
  return (ret & 0xFF);
}

int pcie40_b2l_2ndrxdata(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_BUFDOWN0 + ch*0x20);
  return ((ret & 0xFF00) >> 8);
}

int pcie40_b2l_1strxdata(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_BUFDOWN1 + ch*0x20);
  return (ret & 0xFF);
}

int pcie40_b2l_3rdcharisk(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_BUFDOWN1 + ch*0x20);
  return ((ret & 0x300) >> 8);
}

int pcie40_b2l_2ndcharisk(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_BUFDOWN1 + ch*0x20);
  return ((ret & 0xC00) >> 10);
}

int pcie40_b2l_1stcharisk(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_BUFDOWN1 + ch*0x20);
  return ((ret & 0x3000) >> 12);
}

int pcie40_b2l_cnt_linkdown(int dev, int ch){
  int ret = -1;
  ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_CNTDOWN + ch*0x20);
  
  return ( ret & 0xFFFF);
}

int pcie40_b2l_monstatus(int dev, int ch){
  int ret = -1;
  return ret = ecs_read(dev, SLC_BAR,  SLC_B2LINK_MONSTATUS + ch*0x20); 
}

int pcie40_b2l_vaild_channel(int dev, int ch){
  return B(pcie40_b2l_monstatus(dev, ch), 1);
}

int pcie40_b2l_usebadidle(int dev, int ch){
  return B(pcie40_b2l_monstatus(dev, ch), 2);
}

int pcie40_b2l_usebaddata(int dev, int ch){
  return B(pcie40_b2l_monstatus(dev, ch), 3);
}

int pcie40_b2l_sta_linkdown(int dev, int ch){
  return B(pcie40_b2l_monstatus(dev, ch), 4);

}








